
#
# Shiny App comparing top-score and Pareto optima in rPref
# ********************************************************
#
# Server File
#
# Press "Run App" in RStudio to start the App
#


library(shiny)
library(rPref)
library(ggplot2)
library(grid)


N <- 20 # fixed number of tuples
pref <- high(x) * high(y) # fixed preference

shinyServer(function(input, output, session) {
  
  
  dataset_generator <- eventReactive(input$recalc, {
    
    # Data generator for 2-dim cor/anticor data
    gen_data <- function(N, cor) {  
      rndvals <- matrix(runif(2 * N), N, 2)
      corvals <- matrix(runif(N), N, 2)
      if (cor < 0) { 
        corvals[,2] <- 1 - corvals[,1]
      }
      data <- as.data.frame((1 - abs(cor)) * rndvals + abs(cor) * corvals)
      names(data) <- c("x","y")
      return(data)
    }
    
    # Get input correlation value
    cor <- input$cor
    
    df <- gen_data(N, cor)
    df_max <- psel(df, pref)
    steps <- rbind(df_max, data.frame(x = c(max(df_max$x),0),
                                      y = c(0,max(df$y))) )
    
    list(df = df, df_max = df_max, steps = steps)
    
    
  }, ignoreNULL = FALSE) # ignoreNULL causes recalculation at start
  
  
  observeEvent(input$top_decrease, {
    if (input$top > 0) 
      updateSliderInput(session, "top", value = input$top - 1)
  })
  
  observeEvent(input$top_increase, {
    if (input$top < N) 
      updateSliderInput(session, "top", value = input$top + 1)
  })
  
  
  output$skylinePlot <- renderPlot({
    
    # Get input
    # =========
    
    # read data
    data_input <- dataset_generator()
    df     <- data_input$df
    df_max <- data_input$df_max
    steps  <- data_input$steps
    
    # read parameters
    A <- input$A
    B <- input$B
    top <- input$top
   
   
    # Calculation/Plotting
    # ====================
    
    # Preference / Top Score Selection
    # --------------------------------
    
    # Preference selection
    df_max <- psel(df, pref)
    
    if (top > 0) {
    
      # optimal tuples
      rnk <- rank(with(df, -(A * x + B * y)))
      top_points <- df[rnk <= top,]
      
      # calculate number of top points in optima
      top_in_optima <- intersect(rownames(top_points),rownames(df_max))
      
      # ciritical point
      crit_point <- df[rnk == top,]
      
      # Calculate ascent
      asc <- -A/B
      
      # Calculate border line for scoring top-k
      t <- crit_point$y - asc * crit_point$x
      
      # Left point, set either x=0 or y=1
      if (t <= 1) {
        linedf <- data.frame(x=0, y=t)
      } else {
        linedf <- data.frame(x=(1-t)/asc, y=1)
      }
      
      # Right point, set either x=1 or y=0
      if (asc+t >= 0) {
        linedf <- rbind(linedf, data.frame(x=1, y=asc+t))
      } else {
        linedf <- rbind(linedf, data.frame(x=-t/asc, y=0))
      }
    }
    
    
    # create feedback str
    # -------------------
    
    feedback <- paste0(nrow(df_max), " points are Pareto-optimal")
    
    if (top > 0) {
      feedback <- paste0(feedback, " of which ", length(top_in_optima), 
                             " points are in the top-", top, " scored values")
    }
    
    # helper functions for plotting
    # -----------------------------
    
    # calculate arrow for optimization direction
    get_arrow <- function(linedf) {
      asc2 <- -1/asc
      dx <- (0.1^2 / (1 + asc2^2))^0.5
      x0 <- mean(linedf$x)
      y0 <- mean(linedf$y)
      t2 <- y0 - asc2 * x0
      x1 <- x0 + dx
      y1 <- asc2 * x1 + t2
      
      # return arrow
      return(data.frame(x0 = x0, x1 = x1, y0 = y0, y1 = y1))
    }
    
    # get a mapping such that all characteristics can be manually defined
    get_mapping <- function(str, m = aes(x = x, y = y)) {
      m$colour <- str
      m$size <- str
      m$linetype <- str
      m$shape <- str
      return(m)
    }
    
    # Do the plotting
    # ---------------
    
    # Mappings for Pareto front
    m_opt   <- get_mapping("Optimal tuples")
    m_front <- get_mapping("Pareto front line")
    
    # calculate front line
    steps <- rbind(df_max, data.frame(x=c(max(df_max$x),0),
                                   y=c(0,max(df$y))))
    
    # Plot tuples, optimal tuples and front line
    m <- aes(x = x, y = y)
    p <- ggplot() + geom_point(m, df, cex=3, color="darkgrey") +
         coord_equal(ratio=1, xlim = c(-0.05, 1.05), ylim = c(-0.05, 1.05)) +
         geom_point(m_opt, df_max) + geom_step(m_front, steps, direction="vh") 
         
    if (top > 0) {
      
      # Add top points
      m_score <- get_mapping("Top score tuples")
      p <- p + geom_point(m_score, top_points)
      
      # Add sep line with arrow
      m_sep <- get_mapping("Separation line")
      m_arr <- get_mapping("Separation line", aes(x = x0, xend = x1, y = y0, yend = y1))
      p <- p + geom_line(m_sep, linedf) +
               geom_segment(m_arr, get_arrow(linedf), arrow = arrow())
      
      # Add legend
      leg_name <- "Points and Lines"
      p <- p + scale_color_manual(   values = c("red",   "red",   "darkgreen", "blue"),  name = leg_name) + 
               scale_linetype_manual(values = c("blank", "solid", "solid",     "blank"), name = leg_name) + 
               scale_size_manual(    values = c(3,       0.5,       1,           4),       name = leg_name) + 
               scale_shape_manual(   values = c(16,      1,         1,           0),       name = leg_name)
      
    } else {
      
      # Add legend
      leg_name <- "Points and Lines"
      p <- p + scale_color_manual(   values = c("red",   "red"),   name = leg_name) + 
               scale_linetype_manual(values = c("blank", "solid"), name = leg_name) + 
               scale_size_manual(    values = c(3,       0.5),     name = leg_name) + 
               scale_shape_manual(   values = c(16,      1),       name = leg_name)
      
    }
    
    
    # Create output
    # =============
    
    # return feedback
    output$state <- renderText({feedback})
    
    # Finally plot the graph
    p
  })
})